/* azzuP2 simulator - join the fantasy processor club
 * Now uses ncurses to get proper uncooked input.
 * Adam Sampson, Feb '98
 */

int debug=0;
#define DEBUG (debug == 1)

#define DATATYPE unsigned short int
#define DATABYTES 2
#define DATAMAXVAL 0xFFFF

DATATYPE mem[DATAMAXVAL + 1];
#define PC mem[DATAMAXVAL]
#define REGA mem[DATAMAXVAL - 1]
#define REGB mem[DATAMAXVAL - 2]
#define COND mem[DATAMAXVAL - 3]
#define INTPC mem[DATAMAXVAL - 4]
#define INTCOUNT mem[DATAMAXVAL - 5]
#define INTVEC mem[DATAMAXVAL - 6]
#define SP mem[DATAMAXVAL - 7]

#include <stdio.h>
#include <curses.h>

/* Prototypes */
int emulate();
int io_out(DATATYPE addr, DATATYPE value);
DATATYPE io_in(DATATYPE addr);

int main(int argc, char **argv) {
   FILE *infile;
   int dum;
   
   if (argc < 2) {
      printf("usage: sim [-d] <memimage>\n");
      exit(20);
   };
   if (strcmp(argv[1], "-d") == 0) debug = 1;
   if ((infile = fopen(argv[argc-1], "r")) == NULL) {
      printf("couldn't open image file\n");
      exit(20);
   };
   
   for (dum = 0; dum <= DATAMAXVAL; dum++) mem[dum] = 0;
   
   for (PC = 0; !feof(infile); PC++) {
      for (dum = 0; dum < DATABYTES; dum++) {
	mem[PC] = (mem[PC] << 8) + fgetc(infile);
      };
      if DEBUG fprintf(stderr, "%x: %x\n", PC, mem[PC]);
   };
   
   emulate();
   
   fclose(infile);
};

int emulate() {
   int running = 1;
   DATATYPE inst, src, dest, value, cond;
   
   /* Initialise the ncurses stuff. */
   initscr(); cbreak(); noecho(); nonl(); nodelay(stdscr, TRUE);
   intrflush(stdscr, FALSE); keypad(stdscr, TRUE);
   scrollok(stdscr, TRUE); idlok(stdscr, TRUE);
   
   if DEBUG fprintf(stderr, "emulation start\n");
   PC = 0;
   INTCOUNT = 0;
   
   while (running) {
      inst = mem[PC];
      if DEBUG fprintf(stderr, "PC:%x inst:%x A:%x B:%x C:%x IC:%x SP:%x\n", PC, inst, REGA, REGB, COND, INTCOUNT, SP);
      PC++;
      if (inst < 0x80) {
	 /* handle a mv */
	 src = mem[PC]; PC++;
	 dest = mem[PC]; PC++;
	 cond = inst & 0x70;
	 if DEBUG fprintf(stderr, "mv %x:%x %x:%x cond:%x\n", inst&0x0c, src, inst&0x03, dest, cond);
	 if ( (cond==0x00)
	     || ((cond==0x20) && (COND==0))
	     || ((cond==0x30) && (COND!=0))
	     || ((cond==0x40) && (((COND<<1)&0x01)==0))
	     || ((cond==0x50) && (((COND<<1)&0x01)!=0)) ) {
	    switch (inst & 0x0C) {
	     case 0: value = src; break;
	     case 4: value = mem[src]; break;
	     case 8: value = mem[mem[src]]; break;
	     case 12: value = io_in(src); break;
	    };
	    switch (inst & 0x03) {
	     case 0:
	       if DEBUG fprintf(stderr, "write to literal???\n"); 
	       break;
	     case 1: mem[dest] = value; break;
	     case 2: mem[mem[dest]] = value; break;
	     case 3: io_out(dest, value); break;
	    };
	 };
      } else {
	 /* handle a do */
	 switch(inst & 0x3F) {
	  case 0x00: REGA = 0; break;
	  case 0x01: REGA = REGA + REGB; break;
	  case 0x02: REGA = REGA - REGB; break;
	  case 0x03: REGA = !REGA; break;
	  case 0x04: REGA = REGA & REGB; break;
	  case 0x05: REGA = REGA | REGB; break;
	  case 0x06: mem[SP] = PC; SP++; PC = REGA; break;
	  case 0x07: SP--; PC = mem[SP]; break;
	  case 0x3F: running = 0; break;
	  default:
	    if DEBUG fprintf(stderr, "unknown do instruction: %x\n", inst);
	    break;
	 };
      };
      /* Handle the counted interrupts "feature".
       * This is suprisingly short for all that it does. 
       */
      if (INTCOUNT > 0) {
	 INTCOUNT--;
	 if (INTCOUNT == 0) {
	    if DEBUG fprintf(stderr, "interrupt! int to %x\n", INTVEC);
	    INTPC = PC;
	    PC = INTVEC;
	 };
      };
   };
   if DEBUG fprintf(stderr, "emulation stop, PC=%x\n", PC);
   endwin();
};

int io_out(DATATYPE addr, DATATYPE value) {
   switch (addr) {
    case 0x00:
      if DEBUG fprintf(stderr, "print: %c\n", value);
      echochar((char)value);
      if (value == 13) echochar(10);
      break;
    default:
      if DEBUG fprintf(stderr, "unknown IO write: %x %x\n", addr, value);
      break;
   };
};
   
DATATYPE io_in(DATATYPE addr) {
   int ch;
   
   switch (addr) {
    case 0x00:
      ch = getch();
      if (ch == ERR) {
	 return 0;
      } else {
	 return ch;
      };
      break;
    default:
      if DEBUG fprintf(stderr, "unknown IO read: %x\n", addr);
      break;
   };
};
   
